@@ -1,4 +1,4 @@
-/* $Id: clock.c 29078 2009-01-02 13:21:47Z nick $ */
+/* $Id$ */
 /*
  * Copyright (c) 2007 Nick Schermer <nick@xfce.org>
  *
@@ -35,6 +35,18 @@
 
 #define USE_DEBUG_TIME (0)
 
+#if GLIB_CHECK_VERSION (2, 14, 0)
+#define clock_timeout_add_seconds(interval,function,data) \
+    g_timeout_add_seconds (interval,function,data)
+#define clock_timeout_add_seconds_full(priority,interval,function,data,notify) \
+    g_timeout_add_seconds_full (priority,interval,function,data,notify)
+#else
+#define clock_timeout_add_seconds(interval,function,data) \
+    g_timeout_add ((interval) * 1000,function,data)
+#define clock_timeout_add_seconds_full(priority,interval,function,data,notify) \
+    g_timeout_add_full (priority,(interval) * 1000,function,data,notify)
+#endif
+
 
 
 /** prototypes **/
@@ -43,7 +55,7 @@ static void         xfce_clock_util_get_debug_localtime  (struct tm       *tm);
 #endif
 static guint        xfce_clock_util_interval_from_format (const gchar     *format);
 static guint        xfce_clock_util_next_interval        (guint            timeout_interval);
-static gboolean     xfce_clock_tooltip_update            (gpointer         user_data);
+static void         xfce_clock_tooltip_update            (ClockPlugin     *plugin);
 static gboolean     xfce_clock_tooltip_sync_timeout      (gpointer         user_data);
 static gboolean     xfce_clock_widget_sync_timeout       (gpointer         user_data);
 static ClockPlugin *xfce_clock_plugin_init               (XfcePanelPlugin *panel_plugin);
@@ -63,7 +75,8 @@ XFCE_PANEL_PLUGIN_REGISTER_INTERNAL (xfce_clock_plugin_construct);
 void
 xfce_clock_util_get_localtime (struct tm *tm)
 {
-    time_t now = time (0);
+    time_t now;
+    time (&now);
 
 #ifndef HAVE_LOCALTIME_R
     struct tm *tmbuf;
@@ -121,10 +134,9 @@ static guint
 xfce_clock_util_interval_from_format (const gchar *format)
 {
     const gchar *p;
-    guint        interval = CLOCK_INTERVAL_HOUR;
 
     if (G_UNLIKELY (format == NULL))
-        return CLOCK_INTERVAL_HOUR;
+        return CLOCK_INTERVAL_MINUTE;
 
     for (p = format; *p != '\0'; ++p)
     {
@@ -140,16 +152,11 @@ xfce_clock_util_interval_from_format (const gchar *format)
                 case 'T':
                 case 'X':
                     return CLOCK_INTERVAL_SECOND;
-
-                case 'M':
-                case 'R':
-                    interval = CLOCK_INTERVAL_MINUTE;
-                    break;
             }
         }
     }
 
-    return interval;
+    return CLOCK_INTERVAL_MINUTE;
 }
 
 
@@ -158,37 +165,16 @@ static guint
 xfce_clock_util_next_interval (guint timeout_interval)
 {
     struct tm tm;
-    GTimeVal  timeval;
-    guint     interval;
-
-    /* get the precise time */
-    g_get_current_time (&timeval);
-
-    /* ms to next second */
-    interval = 1000 - (timeval.tv_usec / 1000);
-
-    /* get current time */
-    xfce_clock_util_get_localtime (&tm);
 
     /* add the interval time to the next update */
-    switch (timeout_interval)
+    if (timeout_interval == CLOCK_INTERVAL_MINUTE)
     {
-        case CLOCK_INTERVAL_HOUR:
-            /* ms to next hour */
-            interval += (60 - tm.tm_min) * CLOCK_INTERVAL_MINUTE;
-
-            /* fall-through to add the minutes */
-
-        case CLOCK_INTERVAL_MINUTE:
-            /* ms to next minute */
-            interval += (60 - tm.tm_sec) * CLOCK_INTERVAL_SECOND;
-            break;
-
-        default:
-            break;
+        /* ms to next minute */
+        xfce_clock_util_get_localtime (&tm);
+        return 60 - tm.tm_sec;
     }
 
-    return interval;
+    return 0;
 }
 
 
@@ -223,10 +209,9 @@ xfce_clock_util_strdup_strftime (const gchar *format,
 
 
 /** tooltip functions **/
-static gboolean
-xfce_clock_tooltip_update (gpointer user_data)
+static void
+xfce_clock_tooltip_update (ClockPlugin *plugin)
 {
-    ClockPlugin        *plugin = (ClockPlugin *) user_data;
     gchar              *string;
     struct tm           tm;
 #if !GTK_CHECK_VERSION (2,12,0)
@@ -238,7 +223,7 @@ xfce_clock_tooltip_update (gpointer user_data)
 #endif
 
     if (G_UNLIKELY (plugin->tooltip_format == NULL))
-        return TRUE;
+        return;
 
     /* get the local time */
     xfce_clock_util_get_localtime (&tm);
@@ -256,8 +241,41 @@ xfce_clock_tooltip_update (gpointer user_data)
 
     /* cleanup */
     g_free (string);
+}
 
-    return TRUE;
+
+
+static gboolean
+xfce_clock_tooltip_timeout (gpointer user_data)
+{
+    ClockPlugin *plugin = (ClockPlugin *) user_data;
+    struct tm    tm;
+
+    xfce_clock_tooltip_update (plugin);
+
+    if (plugin->tooltip_interval == CLOCK_INTERVAL_MINUTE)
+    {
+        xfce_clock_util_get_localtime (&tm);
+        plugin->tooltip_restart = tm.tm_sec != 0;
+    }
+
+    return !plugin->tooltip_restart;
+}
+
+
+
+static void
+xfce_clock_tooltip_timeout_destroyed (gpointer user_data)
+{
+    ClockPlugin *plugin = (ClockPlugin *) user_data;
+
+    plugin->tooltip_timeout_id = 0;
+
+    if (plugin->tooltip_restart)
+    {
+        plugin->tooltip_restart = FALSE;
+        xfce_clock_tooltip_sync (plugin);
+    }
 }
 
 
@@ -268,7 +286,10 @@ xfce_clock_tooltip_sync_timeout (gpointer user_data)
     ClockPlugin *plugin = (ClockPlugin *) user_data;
 
     /* start the tooltip update interval */
-    plugin->clock_timeout_id = g_timeout_add (plugin->tooltip_interval, xfce_clock_tooltip_update, plugin);
+    plugin->tooltip_timeout_id = clock_timeout_add_seconds_full (G_PRIORITY_DEFAULT_IDLE,
+                                                                 plugin->tooltip_interval,
+                                                                 xfce_clock_tooltip_timeout, plugin,
+                                                                 xfce_clock_tooltip_timeout_destroyed);
 
     /* manual update for this timeout */
     xfce_clock_tooltip_update (plugin);
@@ -297,8 +318,19 @@ xfce_clock_tooltip_sync (ClockPlugin *plugin)
     /* get the interval to the next update */
     interval = xfce_clock_util_next_interval (plugin->tooltip_interval);
 
-    /* start the sync timeout */
-    plugin->tooltip_timeout_id = g_timeout_add (interval, xfce_clock_tooltip_sync_timeout, plugin);
+    if (interval > 0 && plugin->tooltip_interval != CLOCK_INTERVAL_SECOND)
+    {
+         /* start the sync timeout */
+         plugin->tooltip_timeout_id = clock_timeout_add_seconds (interval, xfce_clock_tooltip_sync_timeout, plugin);
+     }
+     else
+     {
+         /* start the real timeout */
+         plugin->tooltip_timeout_id = clock_timeout_add_seconds_full (G_PRIORITY_DEFAULT_IDLE,
+                                                                      plugin->tooltip_interval,
+                                                                      xfce_clock_tooltip_timeout, plugin,
+                                                                      xfce_clock_tooltip_timeout_destroyed);
+     }
 
     /* update the tooltip */
     xfce_clock_tooltip_update (plugin);
@@ -307,6 +339,43 @@ xfce_clock_tooltip_sync (ClockPlugin *plugin)
 
 
 /** clock widget functions **/
+static gboolean
+xfce_clock_widget_timeout (gpointer user_data)
+{
+    ClockPlugin *plugin = (ClockPlugin *) user_data;
+    gboolean     result;
+    struct tm    tm;
+
+    /* update the widget */
+    result = (plugin->update) (plugin->widget);
+
+    if (result && plugin->interval == CLOCK_INTERVAL_MINUTE)
+    {
+        xfce_clock_util_get_localtime (&tm);
+        plugin->restart = tm.tm_sec != 0;
+    }
+
+    return result && !plugin->restart;
+}
+
+
+
+static void
+xfce_clock_widget_timeout_destroyed (gpointer user_data)
+{
+    ClockPlugin *plugin = (ClockPlugin *) user_data;
+
+    plugin->clock_timeout_id = 0;
+
+    if (plugin->restart)
+    {
+        plugin->restart = FALSE;
+        xfce_clock_widget_sync (plugin);
+    }
+}
+
+
+
 static gboolean
 xfce_clock_widget_sync_timeout (gpointer user_data)
 {
@@ -315,7 +384,10 @@ xfce_clock_widget_sync_timeout (gpointer user_data)
     if (G_LIKELY (plugin->widget))
     {
         /* start the clock update timeout */
-        plugin->clock_timeout_id = g_timeout_add (plugin->interval, plugin->update, plugin->widget);
+        plugin->clock_timeout_id = clock_timeout_add_seconds_full (G_PRIORITY_DEFAULT_IDLE,
+                                                                   plugin->interval,
+                                                                   xfce_clock_widget_timeout, plugin,
+                                                                   xfce_clock_widget_timeout_destroyed);
 
         /* manual update for this interval */
         (plugin->update) (plugin->widget);
@@ -349,8 +421,19 @@ xfce_clock_widget_sync (ClockPlugin *plugin)
         /* get the interval to the next update */
         interval = xfce_clock_util_next_interval (plugin->interval);
 
-        /* start the sync timeout */
-        plugin->clock_timeout_id = g_timeout_add (interval, xfce_clock_widget_sync_timeout, plugin);
+        if (interval > 0 && plugin->interval != CLOCK_INTERVAL_SECOND)
+        {
+            /* start the sync timeout */
+            plugin->clock_timeout_id = clock_timeout_add_seconds (interval, xfce_clock_widget_sync_timeout, plugin);
+        }
+        else
+        {
+            /* start the real timeout */
+            plugin->clock_timeout_id = clock_timeout_add_seconds_full (G_PRIORITY_DEFAULT_IDLE,
+                                                                       plugin->interval,
+                                                                       xfce_clock_widget_timeout, plugin,
+                                                                       xfce_clock_widget_timeout_destroyed);
+        }
     }
 }
 
